#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>



static void freeGraph(int numVertices, int * * graph)
{
  if (graph != NULL)
  {
    int i;
    for (i = 0; i < numVertices; ++i)
    {
      if (graph[i] != NULL)
      {
        free(graph[i]);
        graph[i] = NULL;
      }
    }
    free(graph);
  }
}



static int * * initGraph(int numVertices)
{
  int i;
  int * * graph = (int * *)malloc(sizeof(int *) * numVertices);
  if (graph == NULL)
  {
    return NULL;
  }
  
  for (i = 0; i < numVertices; ++i)
  {
    graph[i] = NULL;
  }
  
  for (i = 0; i < numVertices; ++i)
  {
    graph[i] = (int *)malloc(sizeof(int) * numVertices);
    if (graph[i] == NULL)
    {
      freeGraph(numVertices, graph);
      return NULL;
    }
    else
    {
      int j;
      for (j = 0; j < numVertices; ++j)
      {
        graph[i][j] = i == j ? 0 :
#ifndef USE_RANDOM_GRAPH
            (i - j == 1 || j - i == 1 ? 1 : numVertices + 5);
#else
            (rand() & 8191) + 1;
#endif
      }
    }
  }

  return graph;
}



static void floydWarshall(int numVertices, int * * graph)
{
  int k, i, j;
  for (k = 0; k < numVertices; ++k)
  {
    for (i = 0; i < numVertices; ++i)
    {
      for (j = 0; j < numVertices; ++j)
      {
        int pathSum = graph[i][k] + graph[k][j];
        if (graph[i][j] > pathSum)
        {
          graph[i][j] = pathSum;
        }
      }
    }
  }
}



static void printGraph(int numVertices, int * * graph)
{
  int i;
  for (i = 0; i < numVertices; ++i)
  {
    int j;
    printf("%4d", graph[i][0]);
    for (j = 1; j < numVertices; ++j)
    {
      printf(" %4d", graph[i][j]);
    }
    printf("\n");
  }
}



int main(int argc, char * argv[])
{
  int numVertices = 0;
  int showResults = 0;
  int i;
  int * * graph;
  double startTime;
  double endTime;

  MPI_Init(&argc, &argv);
  
  for (i = 1; i < argc; ++i)
  {
    if (strcmp(argv[i], "--show-results") == 0)
    {
      showResults = 1;
    }
    else
    {
      numVertices = atoi(argv[i]);
    }
  }
  
  if (numVertices <= 0)
  {
    fprintf(stderr, "Usage: %s [--show-results] <num_vertices>\n", argv[0]);
    MPI_Finalize();
    return 1;
  }
  
  fprintf(stderr, "Running the Floyd-Warshall algorithm for a graph with %d vertices.\n", numVertices);
  
#ifdef USE_RANDOM_GRAPH
#ifdef USE_RANDOM_SEED
  srand(USE_RANDOM_SEED);
#endif
#endif

  graph = initGraph(numVertices);
  if (graph == NULL)
  {
    fprintf(stderr, "Error initializing the graph for the algorithm.\n");
    MPI_Finalize();
    return 2;
  }

  if (showResults)
  {
    printGraph(numVertices, graph);
  }
  
  startTime = MPI_Wtime();
  
  floydWarshall(numVertices, graph);
  
  endTime = MPI_Wtime();
  
  fprintf(
      stderr,
      "The time required for the Floyd-Warshall algorithm on a %d-node graph with %d process(es): %f.\n",
      numVertices,
      1, /* numProcesses */
      endTime - startTime
  );
  
  if (showResults)
  {
    printGraph(numVertices, graph);
  }
  
  freeGraph(numVertices, graph);
  
  MPI_Finalize();

  return 0;
}
